/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#ifndef MPEGFILE_H
#define MPEGFILE_H

// for reading file size
#include <sys/types.h>
#include <sys/stat.h>

// because of definitions in header
#include "header.H"

#ifdef __cplusplus

#include <algorithm>
#include <cstdio>
#include <cmath>
#include <iostream>
#include <string>

#include "allLayers.H"

using namespace std;


//--------------------------------------------------------------------------

// Class to open mpeg files, checking the file type
// and creating audio object of according type (layer I/II/III)
// It inherits all functionality to parse frame headers from
// the Header class.
// It also interfaces to required data of the frame content.
class MPEGfile : public Header {

friend class Layer1;
friend class Layer2;
friend class Layer3;

#define MAX_BUFSIZE      1728     // no of bytes for storing a frame
  // mpeg1:
  // LI:   max 448kbit/s / 32kHz * 12 slots = 168 slots => 672 byte
  // LII:  max 384kbit/s / 32kHz * 144 slots = 1728 byte
  // LIII: max 320kbit/s / 32kHz * 144 slots = 1440 byte
  // mpeg2:
  // LI:   max 256kbit/s / 16kHz * 12 slots = 192 slots => 768 byte
  // LII:  max 160kbit/s / 16kHz * 144 slots = 1440 byte
  // LIII: max 160kbit/s / 16kHz * 144 slots = 1440 byte
 
public:
    MPEGfile () : fd(NULL), audio(NULL), lastlayer(RESERVED), currTime(0.0),
                  fileDuration(0.0), frameNo(0) {}
    MPEGfile (string filestr);
    virtual ~MPEGfile ();

    string    file();           // filename
    float     at();             // time position
    float     duration();       // duration of sound file (if calculated)
    long      frame();          // current frame number
    long      frame(float t);   // which framenumber corresponds to this sec?
    float     time(long f);     // which second corresponds to this frame?
    bool      data_available(); // end of data reached?
    bool      seek_time (float t);

    float     frame_duration(); // duration of a frame
    float     scf_duration();   // duration of a scalefactor

    bool      parse_header();
    bool      parse_data(bool decode = true);
    bool      parse_frame(bool decode = true); // header+data
    bool      skip_data();
    bool      skip_frame();  // header+data

    // layer-specific access to audio data
    int   bitallocation    (unsigned int ch, unsigned int sb);
    int   scfsi            (unsigned int ch, unsigned int sb);
    float scalefactor      (unsigned int ch, unsigned int sb, 
			    unsigned int ss=0);
    int   sample           (unsigned int ch, unsigned int sb, 
			    unsigned int no, unsigned int ss=0);
    double restored_sample (unsigned int ch, unsigned int sb, 
			    unsigned int no, unsigned int ss=0);
    short pcm_sample       (unsigned int ch, unsigned int sb, 
		            unsigned int no, unsigned int ss=0);

    // layer-III specific data
    void printSideinfo();
    unsigned int part2_3_length     (unsigned int ch, unsigned int gr);
    unsigned int big_values         (unsigned int ch, unsigned int gr);
    unsigned int count1_values      (unsigned int ch, unsigned int gr);
    unsigned int global_gain        (unsigned int ch, unsigned int gr);
    unsigned int scalefac_compress  (unsigned int ch, unsigned int gr);
    bool         window_switching   (unsigned int ch, unsigned int gr);
    unsigned int blocktype          (unsigned int ch, unsigned int gr);
    bool         mixedblock         (unsigned int ch, unsigned int gr);
    unsigned int table_select       (unsigned int ch, unsigned int gr, 
				     unsigned int area);
    unsigned int subblock_gain      (unsigned int ch, unsigned int gr, 
				     unsigned int area);
    unsigned int region0_samps      (unsigned int ch, unsigned int gr);
    unsigned int region1_samps      (unsigned int ch, unsigned int gr);
    unsigned int region2_samps      (unsigned int ch, unsigned int gr);
    unsigned int preflag            (unsigned int ch, unsigned int gr);
    unsigned int scalefac_scale     (unsigned int ch, unsigned int gr);
    unsigned int count1table_select (unsigned int ch, unsigned int gr);
    unsigned int scf_band_bound_l   (unsigned int subbandindex);
    unsigned int scf_band_bound_s   (unsigned int subbandindex);
    unsigned int slen1              (unsigned int ch, unsigned int gr);
    unsigned int slen2              (unsigned int ch, unsigned int gr);

protected:
    // audio data stored in bitbuffer
    unsigned int buffer[MAX_BUFSIZE/4];

    unsigned int bitsread; // from buffer
    unsigned int readbitsfrombuffer(unsigned int nobits);

private:
    string   filename;
    
    FILE *fd;
    
    AllLayers *audio;

    // remembers which layer the last frame belonged to
    Layer lastlayer;

    // remeber the current file position and file duration in seconds
    float currTime;
    float fileDuration;

    // remember the current frame number
    long frameNo;

    // private functions
    bool create_layer();
    bool read_checksum();
    bool buffer_data();

};

//--------------------------------------------------------------------------

// inline operators

extern "C" {

#else /* __cplusplus */

typedef struct MPEGfile MPEGfile;

#endif /* __cplusplus */


  MPEGfile * maaateP_new_mpegfile (char * filestr);

  void maaateP_destroy (MPEGfile * mf);
  
  /* filename */
  char *    maaateP_get_filename(MPEGfile * mf);
  
  /* header functions */
  Version      maaateP_version(MPEGfile * mf);
  Layer        maaateP_layer(MPEGfile * mf);
  int          maaateP_crcprotected(MPEGfile * mf);
  int          maaateP_bitrate(MPEGfile * mf);
  int          maaateP_bitrate_per_channel(MPEGfile * mf);
  int          maaateP_bitrate_index(MPEGfile * mf);
  double       maaateP_samplingrate(MPEGfile * mf);
  int          maaateP_samplingrate_index(MPEGfile * mf);
  int          maaateP_padding(MPEGfile * mf);
  int          maaateP_extended(MPEGfile * mf);
  Mode         maaateP_mode(MPEGfile * mf);
  const char*  maaateP_mode_str(MPEGfile * mf);
  int          maaateP_channels(MPEGfile * mf);
  int          maaateP_mode_ext(MPEGfile * mf);
  const char*  maaateP_mode_ext_str(MPEGfile * mf);
  int          maaateP_intensity_stereo_bound(MPEGfile * mf);
  int          maaateP_copyright(MPEGfile * mf);
  int          maaateP_original(MPEGfile * mf);
  int          maaateP_emphasis(MPEGfile * mf);
  int          maaateP_samples_per_frame(MPEGfile * mf);
  unsigned int maaateP_framesize(MPEGfile * mf);
  int          maaateP_subbands(MPEGfile * mf);
  unsigned int maaateP_check_sum(MPEGfile * mf);
  void         maaateP_printheader (MPEGfile * mf);

  /* time position */
  float     maaateP_at(MPEGfile * mf);
  
  /* duration of sound file (if calculated) */
  float     maaateP_get_duration(MPEGfile * mf);
  
  /* current frame number */
  long      maaateP_get_frame_current(MPEGfile * mf);
  
  /* which framenumber corresponds to this sec? */
  long      maaateP_get_frame(MPEGfile * mf, float t);
  
  /* which second corresponds to this frame? */
  float     maaateP_get_time(MPEGfile * mf, long f);
  
  /* end of data reached? */
  int       maaateP_data_available(MPEGfile * mf);
  
  /* seek to specific time */
  int       maaateP_seek_time (MPEGfile * mf, float t);
  
  /* duration of a frame */
  float     maaateP_get_frame_duration(MPEGfile * mf);
  
  /* duration of a scalefactor */
  float     maaateP_get_scf_duration(MPEGfile * mf);
  
  int       maaateP_parse_header(MPEGfile * mf);
  int       maaateP_parse_data(MPEGfile * mf, int decode);
  int       maaateP_parse_frame(MPEGfile * mf, int decode); /* header+data */
  int       maaateP_skip_data(MPEGfile * mf);
  int       maaateP_skip_frame(MPEGfile * mf);  /* header+data */
  
  /* layer-specific access to audio data */
  int   maaateP_bitallocation    (MPEGfile * mf, unsigned int ch,
				  unsigned int sb);
  int   maaateP_scfsi            (MPEGfile * mf, unsigned int ch,
				  unsigned int sb);
  float maaateP_scalefactor      (MPEGfile * mf, unsigned int ch,
				  unsigned int sb, unsigned int ss);
  int   maaateP_sample           (MPEGfile * mf, unsigned int ch,
				  unsigned int sb,
				  unsigned int no, unsigned int ss);
  double maaateP_restored_sample (MPEGfile * mf, unsigned int ch,
				  unsigned int sb, 
				  unsigned int no, unsigned int ss);
  short maaateP_pcm_sample       (MPEGfile * mf, unsigned int ch,
				  unsigned int sb, 
				  unsigned int no, unsigned int ss);
  
  /* layer-III specific data */
  void maaateP_print_sideinfo(MPEGfile * mf);
  unsigned int maaateP_part2_3_length     (MPEGfile * mf, unsigned int ch,
					   unsigned int gr);
  unsigned int maaateP_big_values         (MPEGfile * mf, unsigned int ch,
					   unsigned int gr);
  unsigned int maaateP_count1_values      (MPEGfile * mf, unsigned int ch,
					   unsigned int gr);
  unsigned int maaateP_global_gain        (MPEGfile * mf, unsigned int ch,
					   unsigned int gr);
  unsigned int maaateP_scalefac_compress  (MPEGfile * mf, unsigned int ch,
					   unsigned int gr);
  int          maaateP_window_switching   (MPEGfile * mf, unsigned int ch,
					   unsigned int gr);
  unsigned int maaateP_blocktype          (MPEGfile * mf, unsigned int ch,
					   unsigned int gr);
  int          maaateP_mixedblock         (MPEGfile * mf, unsigned int ch,
					   unsigned int gr);
  unsigned int maaateP_table_select       (MPEGfile * mf, unsigned int ch,
					   unsigned int gr, unsigned int area);
  unsigned int maaateP_subblock_gain      (MPEGfile * mf, unsigned int ch,
					   unsigned int gr, unsigned int area);
  unsigned int maaateP_region0_samps      (MPEGfile * mf, unsigned int ch,
					   unsigned int gr);
  unsigned int maaateP_region1_samps      (MPEGfile * mf, unsigned int ch,
					   unsigned int gr);
  unsigned int maaateP_region2_samps      (MPEGfile * mf, unsigned int ch,
					   unsigned int gr);
  unsigned int maaateP_preflag            (MPEGfile * mf, unsigned int ch,
					   unsigned int gr);
  unsigned int maaateP_scalefac_scale     (MPEGfile * mf, unsigned int ch,
					   unsigned int gr);
  unsigned int maaateP_count1table_select (MPEGfile * mf, unsigned int ch,
					   unsigned int gr);
  unsigned int maaateP_scf_band_bound_l   (MPEGfile * mf,
					   unsigned int subbandindex);
  unsigned int maaateP_scf_band_bound_s   (MPEGfile * mf,
					   unsigned int subbandindex);
  unsigned int maaateP_slen1              (MPEGfile * mf, unsigned int ch,
					   unsigned int gr);
  unsigned int maaateP_slen2              (MPEGfile * mf, unsigned int ch,
					   unsigned int gr);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
